#ifndef PCIE40_B2DMA_H
#define PCIE40_B2DMA_H
// Header file for Belle 2 ECS functions with PCIe40

// Constants
// BAR number for Slow control interface
#define DMA_BAR 0
#define LLI_BAR 2 

// Address of the register to reset the write FIFO and bit to use
#define DMA_CTRL 0x200
#define DMA_TRIGGER_BIT 0 

#define DMA_RST 0x400
#define DMA_RST_BIT 0
#define DMA_BUSY_LEVEL 0x40C

#define DMA_FIFO_0 0xC00
#define DMA_FIFO_1 0xC04

#define REG_NB_WORD_IN_EVENT 0x505E0
#define REG_NB_EVENTS 0x50600
#define REG_TRIG_FREE_RESET  0x50620
#define DISABLE_GENERATOR_BIT 10
#define USE_DATA_FROM_FIBERS_BIT 9 
#define USE_FREE_BIT 8
#define USE_EXT_TRIG_BIT 11

/// Open DMA
int dma_open( int fd ) ;

/// Open DMA
int dma_close( int fd ) ;

/// Initialize DMA
int dma_initialize( int fd ) ;

/// reset spages addresses
int pcie40_resetSPages( int fd ) ; 

/// Start/Stop DMA
int pcie40_dmaStart( int fd ) ;
int pcie40_dmaStop( int fd ) ;

/// Event size (for internal data generator)
int pcie40_setNbWordInEvent( int fd , unsigned int n ) ;
unsigned int pcie40_getNbWordInEvent( int fd ) ;

/// Number of events (internal data generator)
int pcie40_setNbEvents( int fd , unsigned int n ) ;
unsigned int pcie40_getNbEvents( int fd ) ;

/// Enable/disable internal data generator
int pcie40_enableGenerator( int fd ) ;
int pcie40_disableGenerator( int fd ) ;

/// Use data from fibers
int pcie40_useDataFromFibers( int fd ) ;

/// Use data from generator 
int pcie40_useDataFromGenerator( int fd ) ; 

/// Use free signal
int pcie40_useFreeSignal( int fd ) ;
int pcie40_ignoreFreeSignal( int fd ) ;

/// DMA reset
int pcie40_dmaReset( int fd ) ; 

/// set DMA readout 
int pcie40_dmaSetReadout( int fd ) ;

/// Get a pointer to super page
unsigned int * pcie40_getSuperPagePointer( int link , int spage , int slot ) ;

/// Copy super page memory
unsigned int * pcie40_getSuperPageCopy( int link , int spage , int slot ) ;

/// Free super page when it is ready
int pcie40_freeSuperPage( int fd , int spage ) ;

/// Set trigger frequency of data generator
int pcie40_setGeneratorTriggerFrequency( int fd , unsigned int freq ) ;

/// Get trigger frequency of data generator
int pcie40_getGeneratorTriggerFrequency( int fd ) ;

/// Use external trigger (b2tt) or internal for the generator
int pcie40_enableExternalTrigger( int fd ) ;
int pcie40_disableExternalTrigger( int fd ) ;

/// Set DMA size from header
int pcie40_setSizeFromHeader( int fd ) ;
int pcie40_disableSizeFromHeader( int fd ) ;

/// Set busy levels
int pcie40_setBusyLevel( int fd , unsigned int level ) ;

/// reset address of DMA buffer
int pcie40_b2dmabufferreset(int fd);

/// read entire DMA buffer
int pcie40_b2dmabufferread(int fd , unsigned int result[]);

/// read pointer to DMA
int pcie40_b2dmapointerread(int fd , unsigned int ** pData );
#endif // PCIE40_B2DMA_H
